﻿using System;
using SDX.Compiler;
using Mono.Cecil;
using Mono.Cecil.Cil;
using System.Linq;
using System.Collections.Generic;

public class RandomSpawnPointPrefab : IPatcherMod
{
    public bool Link(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        InsertPregenerateHubCellDataCoroutine(gameModule, modModule);
        InsertPregenerateHubCellData(gameModule, modModule);
        InsertLoadData(gameModule, modModule);
        InsertAddRandomSpawnPointListHubCell(gameModule, modModule);
        InsertAddRandomSpawnPointListHubCellWilderness(gameModule, modModule);
        InsertCW(gameModule, modModule);
        InsertOW(gameModule, modModule);
        InsertRulesFromXml(gameModule, modModule);
        InsertPlayerMoveOZ(gameModule, modModule);
        InsertPlayerMoveZW(gameModule, modModule);
        InsertGameManagerOQ(gameModule, modModule);
        InsertGameManagerRequestToSpawnPlayer(gameModule, modModule);
        InsertCopyPrefabData(gameModule, modModule);
        InsertGameStageDefinitionRead(gameModule, modModule);
        InsertNetPackageWorldSpawnPointsWrite(gameModule, modModule);
        InsertNetPackageWorldSpawnPointsRead(gameModule, modModule);
        InsertNetPackageWorldSpawnPointsGetLength(gameModule, modModule);
        return true;
    }
    public void InsertNetPackageWorldSpawnPointsWrite(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "NetPackageWorldSpawnPoints");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "write");
        Instruction insertPos = methodDef.Body.Instructions[methodDef.Body.Instructions.Count - 1];

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition setWriteBinaryData = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "SetWriteBinaryData");
        MethodReference setWriteBinaryDataRef = gameModule.Import(setWriteBinaryData);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_1));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, setWriteBinaryDataRef));

    }
    public void InsertNetPackageWorldSpawnPointsRead(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "NetPackageWorldSpawnPoints");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "read");
        Instruction insertPos = methodDef.Body.Instructions[methodDef.Body.Instructions.Count - 1];

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition setReadBinaryData = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "SetReadBinaryData");
        MethodReference setReadBinaryDataRef = gameModule.Import(setReadBinaryData);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_1));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, setReadBinaryDataRef));

    }
    public void InsertNetPackageWorldSpawnPointsGetLength(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "NetPackageWorldSpawnPoints");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "GetLength");
        Instruction insertPos = methodDef.Body.Instructions[methodDef.Body.Instructions.Count - 1];

        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPos)
            {
                jumpList.Add(inst);
            }
        }

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition getBinaryDataSize = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "GetBinaryDataSize");
        MethodReference getBinaryDataSizeRef = gameModule.Import(getBinaryDataSize);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Call, getBinaryDataSizeRef);
        proc1.InsertBefore(insertPos, newJumpTo);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Add));

        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }

    }

    public void InsertGameStageDefinitionRead(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "GameStagesFromXml");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Load");

        TypeDefinition gameStageDef = gameModule.Types.First(d => d.Name == "GameStageDefinition");
        FieldDefinition deadWellnessDownField = gameStageDef.Fields.First(d => d.Name == "DeadWellnessDown");
        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {

            if (inst.OpCode == OpCodes.Stsfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single GameStageDefinition::DaysAliveDifficultyBonus")
            {
                insertPosition = inst.Next;

            }
        }
        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Brfalse_S && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[4]);
        proc1.InsertBefore(insertPosition, newJumpTo);//V_4 is XmlElement

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition setAddAttribute = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "SetAddAttribute");

        MethodReference setAddAttributeRef = gameModule.Import(setAddAttribute);

        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, setAddAttributeRef));

        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }
    }
    public void InsertCopyPrefabData(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == ".ctor" && d.Parameters.Count == 1);
        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = methodDef.Body.Instructions[0];

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition copyPrefabData = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "CopyPrefabData");
        MethodReference copyPrefabDataRef = gameModule.Import(copyPrefabData);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_1));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, copyPrefabDataRef));

    }


    public void InsertGameManagerRequestToSpawnPlayer(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "GameManager");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "RequestToSpawnPlayer");

        Instruction insertPos = null;
        Instruction insertPosRemoveLast = null;

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "SpawnPosition SpawnPointList::GetRandomSpawnPosition(System.Nullable`1<UnityEngine.Vector3>,System.Int32,System.Int32)")
            {
                insertPos = inst.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous;
                insertPosRemoveLast = inst;
            }
        }
        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition getRandomSpawnPositionNoEntityPlayer = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "GetRandomSpawnPositionNoEntityPlayer");
        MethodReference getRandomSpawnPositionNoEntityPlayerRef = gameModule.Import(getRandomSpawnPositionNoEntityPlayer);

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getRandomSpawnPositionNoEntityPlayerRef));

        Instruction remove = insertPosRemoveLast;
        while (remove.Previous != insertPos)
        {
            proc1.Remove(remove.Previous);
        }
        proc1.Remove(insertPos);
        proc1.Remove(insertPosRemoveLast);

    }

    public void InsertGameManagerOQ(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "GameManager");
        TypeDefinition nestedClassDef = classDef.NestedTypes.First(d => d.Name == "ZU");
        MethodDefinition methodDef = nestedClassDef.Methods.First(d => d.Name == "MoveNext");
        FieldDefinition gameManager = nestedClassDef.Fields.First(d => d.Name == "UZ");

        Instruction insertPos = null;
        Instruction insertPosRemoveLast = null;

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "SpawnPosition SpawnPointList::GetRandomSpawnPosition(System.Nullable`1<UnityEngine.Vector3>,System.Int32,System.Int32)")
            {
                insertPos = inst.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous;
                insertPosRemoveLast = inst;
            }
        }
        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition getRandomSpawnPositionNoEntityPlayer = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "GetRandomSpawnPositionNoEntityPlayer");
        MethodReference getRandomSpawnPositionNoEntityPlayerRef = gameModule.Import(getRandomSpawnPositionNoEntityPlayer);

        Instruction newJumpPos = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(insertPos, newJumpPos);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldfld, gameManager));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getRandomSpawnPositionNoEntityPlayerRef));

        Instruction remove = insertPosRemoveLast;
        while (remove.Previous != insertPos)
        {
            proc1.Remove(remove.Previous);
        }
        proc1.Remove(insertPos);
        proc1.Remove(insertPosRemoveLast);

    }
    public void InsertPlayerMoveZW(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "PlayerMoveController");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "ZW");
        FieldDefinition entityPlayerLocal = classDef.Fields.First(d => d.Name == "entityPlayerLocal");
        FieldDefinition gameManager = classDef.Fields.First(d => d.Name == "gameManager");
        FieldDefinition iz = classDef.Fields.First(d => d.Name == "IZ");

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos1 = null;
        Instruction insertPos1RemoveLast = null;
        Instruction insertPos2 = null;
        Instruction insertPos2RemoveLast = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "SpawnPosition SpawnPointList::GetRandomSpawnPosition(System.Nullable`1<UnityEngine.Vector3>,System.Int32,System.Int32)")
            {
                if (inst.Previous.OpCode == OpCodes.Ldc_I4_0)
                {
                    insertPos2 = inst.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous;
                    insertPos2RemoveLast = inst.Next.Next.Next;

                }
                else
                {
                    insertPos1 = inst.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous;
                    insertPos1RemoveLast = inst.Next.Next.Next;

                }
            }

        }

        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && ((Instruction)inst.Operand).Offset == insertPos2.Offset)
            {
                jumpList.Add(inst);
            }
        }

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition getRandomSpawnPositionRange = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "GetRandomSpawnPositionRange");
        MethodReference getRandomSpawnPositionRangeRef = gameModule.Import(getRandomSpawnPositionRange);

        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldfld, entityPlayerLocal));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldfld, gameManager));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Call, getRandomSpawnPositionRangeRef));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Stloc_1));

        Instruction remove = insertPos1RemoveLast;
        while(remove.Previous != insertPos1)
        {
            proc1.Remove(remove.Previous);
        }
        proc1.Remove(insertPos1);
        proc1.Remove(insertPos1RemoveLast);

        MethodDefinition getRandomSpawnPositionNoSet = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "GetRandomSpawnPositionNoSet");
        MethodReference getRandomSpawnPositionNoSetRef = gameModule.Import(getRandomSpawnPositionNoSet);

        Instruction newJumpPos = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(insertPos2, newJumpPos);
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldfld, entityPlayerLocal));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldfld, gameManager));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Call, getRandomSpawnPositionNoSetRef));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Stloc_1));


        remove = insertPos2RemoveLast;
        while (remove.Previous != insertPos2)
        {
            proc1.Remove(remove.Previous);
        }
        proc1.Remove(insertPos2);
        proc1.Remove(insertPos2RemoveLast);

        foreach (Instruction inst in jumpList)
        {
            inst.Operand = newJumpPos;
        }

    }
    public void InsertPlayerMoveOZ(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "PlayerMoveController");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "OZ");
        FieldDefinition entityPlayerLocal = classDef.Fields.First(d => d.Name == "entityPlayerLocal");
        FieldDefinition iz = classDef.Fields.First(d => d.Name == "IZ");

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = null;
        Instruction jumpToPos = null;
        bool isSetInsertPos = false;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "SpawnPosition SpawnPointList::GetRandomSpawnPosition(System.Nullable`1<UnityEngine.Vector3>,System.Int32,System.Int32)")
            {
                insertPos = inst.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous;
                isSetInsertPos = true;
            }
            if (isSetInsertPos && jumpToPos == null && inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodReference)inst.Operand).FullName == "System.Void PlayerMoveController::KW(RespawnType)")
            {
                jumpToPos = inst.Previous.Previous.Previous;
            }

        }

        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if(inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && ((Instruction)inst.Operand).Offset == insertPos.Offset)
            {
                jumpList.Add(inst);
            }
        }


        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition getRandomSpawnPosition = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "GetRandomSpawnPosition");
        MethodReference getRandomSpawnPositionRef = gameModule.Import(getRandomSpawnPosition);

        TypeDefinition spawnPosition = gameModule.Types.First(d => d.Name == "SpawnPosition");
        FieldDefinition spawnPositionUndef = spawnPosition.Fields.First(d => d.Name == "Undef");
        MethodDefinition spawnPositionEqual = spawnPosition.Methods.First(d => d.Name == "Equals");

        Instruction newJumpPos = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(insertPos, newJumpPos);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldfld, entityPlayerLocal));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getRandomSpawnPositionRef));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Stfld, iz));

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldflda, iz));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldsfld, spawnPositionUndef));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Box, spawnPosition));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, spawnPositionEqual));

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Brfalse, jumpToPos));

        foreach(Instruction inst in jumpList)
        {
            inst.Operand = newJumpPos;
        }

    }
    public void InsertCW(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "CW");

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition getRandomSpawnProperty = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "GetRandomSpawnProperty");
        MethodReference getRandomSpawnPropertyRef = gameModule.Import(getRandomSpawnProperty);


        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        Instruction insertPos = methodDef.Body.Instructions[0];

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getRandomSpawnPropertyRef));

    }

    public void InsertOW(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "OW");

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition getRandomSpawnDataToProperty = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "GetRandomSpawnDataToProperty");
        MethodReference getRandomSpawnDataToPropertyRef = gameModule.Import(getRandomSpawnDataToProperty);


        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        Instruction insertPos = methodDef.Body.Instructions[0];

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getRandomSpawnDataToPropertyRef));

    }

    public void InsertRulesFromXml(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition entityClass = gameModule.Types.First(d => d.Name == "RulesFromXml" && d.Namespace == "RWG2.Rules");
        MethodDefinition kz = entityClass.Methods.First(d => d.Name == "KZ");

        int state = 0;
        Instruction jumpBr = null;
        Instruction insertPos = null;
        List<Instruction> jumpBrList = new List<Instruction>();
        foreach (Instruction inst in kz.Body.Instructions)
        {
            switch (state)
            {
                case 0:
                    if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "CellOffset")
                    {
                        state = 1;

                    }
                    break;
                case 1:
                    if (inst.OpCode == OpCodes.Brfalse_S)
                    {
                        jumpBr = inst;
                        insertPos = (Instruction)jumpBr.Operand;
                        jumpBrList.Add(jumpBr);
                        state = 2;
                    }
                    break;
                case 2:
                    if (inst.OpCode == OpCodes.Brfalse_S && ((Instruction)inst.Operand).Offset == insertPos.Offset)
                    {
                        jumpBrList.Add(inst);
                    }
                    break;
            }
        }


        ILProcessor proc1 = kz.Body.GetILProcessor();
        Instruction newJumpPos = Instruction.Create(OpCodes.Ldloc_3);
        proc1.InsertBefore(insertPos, newJumpPos);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_2));





        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition rulesFromXml = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "RulesFromXml");
        MethodReference rulesFromXmlRef = gameModule.Import(rulesFromXml);


        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, rulesFromXmlRef));

        foreach (Instruction inst in jumpBrList)
        {

            inst.Operand = newJumpPos;
        }

    }

    
    public void InsertAddRandomSpawnPointListHubCellWilderness(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HubCell" && d.Namespace == "RWG2");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "YZ");

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = null;
        VariableDefinition vd = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Void System.Collections.Generic.List`1<RWG2.HubCell/Lot>::Add(!0)")
            {
                if (inst.Previous.OpCode == OpCodes.Ldloc_S && inst.Previous.Operand.GetType() == typeof(VariableDefinition))
                {
                    vd = (VariableDefinition)inst.Previous.Operand;
                    insertPos = inst.Next;

                }
            }
        }

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition addRandomSpawnPointListHubCell = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "AddRandomSpawnPointListHubCell");
        MethodReference addRandomSpawnPointListHubCellRef = gameModule.Import(addRandomSpawnPointListHubCell);

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_S, vd));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, addRandomSpawnPointListHubCellRef));

    }

    public void InsertAddRandomSpawnPointListHubCell(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HGSockets" && d.Namespace == "RWG2.HubGenerators");


        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "GenerateLots");


        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos1 = null;
        Instruction insertPos2 = null;
        VariableDefinition vd1 = null;
        VariableDefinition vd2 = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Void System.Collections.Generic.List`1<RWG2.HubCell/Lot>::Add(!0)")
            {
                if (inst.Previous.OpCode == OpCodes.Ldloc_S && inst.Previous.Operand.GetType() == typeof(VariableDefinition))
                {
                    VariableDefinition vd = (VariableDefinition)inst.Previous.Operand;
                    if (vd.Index == 29)
                    {
                        insertPos1 = inst.Next;
                        vd1 = vd;
                    }
                    else
                    {
                        insertPos2 = inst.Next;
                        vd2 = vd;

                    }

                }
            }
        }

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition addRandomSpawnPointListHubCell = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "AddRandomSpawnPointListHubCell");
        MethodReference addRandomSpawnPointListHubCellRef = gameModule.Import(addRandomSpawnPointListHubCell);

        TypeDefinition hbabst = gameModule.Types.First(d => d.Name == "HGAbstract" && d.Namespace == "RWG2.HubGenerators");
        FieldDefinition hubCell = hbabst.Fields.First(d => d.Name == "hubCell");


        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldloc_S, vd1));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldfld, hubCell));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Call, addRandomSpawnPointListHubCellRef));

        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldloc_S, vd2));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldfld, hubCell));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Call, addRandomSpawnPointListHubCellRef));




    }

    public void InsertPregenerateHubCellData(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition baseRWGDataLoader = gameModule.Types.First(d => d.Name == "BaseRWGDataLoader" && d.Namespace == "RWG2");
        MethodDefinition pregenerateHubCellData = baseRWGDataLoader.Methods.First(d => d.Name == "PregenerateHubCellData");

        ILProcessor proc1 = pregenerateHubCellData.Body.GetILProcessor();
        Instruction insertPos1 = null;
        Instruction insertPos2 = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Void Log::Out(System.String)")
            {
                if (inst.Next.OpCode == OpCodes.Ldc_I4_1)
                {
                    insertPos1 = inst.Next;
                }
                else
                {

                }

            }
            if (inst.OpCode == OpCodes.Blt_S)
            {
                if (inst.Previous.OpCode == OpCodes.Ldfld && inst.Previous.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Previous.Operand).FullName == "System.Int32 RWG2.BaseRWGDataLoader::HZ")
                {
                    if (inst.Previous.Previous.Previous.Previous.OpCode == OpCodes.Stloc_1)
                    {
                        insertPos2 = inst.Next;
                    }
                }
            }
        }
        List<Instruction> jumpInstList = new List<Instruction>();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction))
            {
                if (insertPos2.Offset == ((Instruction)inst.Operand).Offset)
                {

                    jumpInstList.Add(inst);
                }
            }
        }


        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");

        MethodDefinition saveData = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "SaveData");
        MethodReference saveDataRef = gameModule.Import(saveData);
        Instruction newJumpTo = Instruction.Create(OpCodes.Call, saveDataRef);
        proc1.InsertBefore(insertPos2, newJumpTo);

        foreach (Instruction inst in jumpInstList)
        {
            inst.Operand = newJumpTo;
        }

    }

    public void InsertPregenerateHubCellDataCoroutine(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition baseRWGDataLoader = gameModule.Types.First(d => d.Name == "BaseRWGDataLoader" && d.Namespace == "RWG2");
        TypeDefinition nestedClass = baseRWGDataLoader.NestedTypes.First(d => d.Name == "DH");
        MethodDefinition methodDef = nestedClass.Methods.First(d => d.Name == "MoveNext");

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = null;
        bool isFirst = true;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Void Log::Out(System.String)")
            {
                if (isFirst)
                {
                    isFirst = false;
                }
                else
                {
                    insertPos = inst.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous;
                }

            }
        }

        List<Instruction> jumpInstList = new List<Instruction>();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction))
            {
                if (insertPos.Offset == ((Instruction)inst.Operand).Offset)
                {

                    jumpInstList.Add(inst);
                }
            }
        }

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition saveData = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "SaveData");
        MethodReference saveDataRef = gameModule.Import(saveData);
        Instruction newJumpTo = Instruction.Create(OpCodes.Call, saveDataRef);
        proc1.InsertBefore(insertPos, newJumpTo);

        foreach (Instruction inst in jumpInstList)
        {
            inst.Operand = newJumpTo;
        }

    }


    public void InsertLoadData(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition world = gameModule.Types.First(d => d.Name == "World");
        MethodDefinition loadWorld = world.Methods.First(d => d.Name == "LoadWorld");

        ILProcessor proc1 = loadWorld.Body.GetILProcessor();

        Instruction insertPos = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Void WorldCreationData::Apply(World,WorldState)")
            {
                insertPos = inst.Previous.Previous.Previous.Previous.Previous;

            }

        }
        List<Instruction> jumpInstList = new List<Instruction>();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction))
            {
                if (insertPos.Offset == ((Instruction)inst.Operand).Offset)
                {

                    jumpInstList.Add(inst);
                }
            }
        }

        TypeDefinition randomSpawnPointPrefabMethod = modModule.Types.First(d => d.Name == "RandomSpawnPointPrefabMethod");
        MethodDefinition loadData = randomSpawnPointPrefabMethod.Methods.First(d => d.Name == "LoadData");
        MethodReference loadDataRef = gameModule.Import(loadData);

        Instruction newJumpTo = Instruction.Create(OpCodes.Call, loadDataRef);
        proc1.InsertBefore(insertPos, newJumpTo);
        foreach(Instruction inst in jumpInstList)
        {
            inst.Operand = newJumpTo;
        }
    }
    

    public void AddRulesetField(ModuleDefinition module)
    {
        TypeDefinition ruleset = module.Types.First(d => d.Name == "Ruleset" && d.Namespace == "RWG2.Rules");
        TypeReference tr = module.Import(typeof(string));
        AddFieldDefinition(ruleset, "RandamSpawnPrefabFilter", FieldAttributes.Public, tr);

    }

    public void AddPrefabField(ModuleDefinition module)
    {
        TypeReference tr = module.Import(typeof(float));
        TypeDefinition prefab = module.Types.First(d => d.Name == "Prefab");

        AddFieldDefinition(prefab, "RandomSpawnOffsetX", FieldAttributes.Public, tr);
        AddFieldDefinition(prefab, "RandomSpawnOffsetY", FieldAttributes.Public, tr);
        AddFieldDefinition(prefab, "RandomSpawnOffsetZ", FieldAttributes.Public, tr);
        tr = module.Import(typeof(byte));
        AddFieldDefinition(prefab, "RandomSpawnRotation", FieldAttributes.Public, tr);
    }
    public bool Patch(ModuleDefinition module)
    {
        Console.WriteLine("==RandomSpawnPointPrefab Patcher===");
        AddPrefabField(module);
        AddRulesetField(module);
        return true;
    }

    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        classDef.Fields.Add(fd);

    }
    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType, object constantvalue)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        fd.Constant = constantvalue;
        classDef.Fields.Add(fd);

    }
    private void SetFieldToPublic(FieldDefinition field)
    {
        field.IsFamily = false;
        field.IsPrivate = false;
        field.IsPublic = true;
        

    }

}
