﻿using System;
using System.Linq;
using System.Xml;
using System.Collections.Generic;
using UnityEngine;
using System.IO;

public static class RandomSpawnPointPrefabMethod
{

    public static void CopyPrefabData(Prefab src, Prefab dst)
    {
        dst.RandomSpawnOffsetX = src.RandomSpawnOffsetX;
        dst.RandomSpawnOffsetY = src.RandomSpawnOffsetY;
        dst.RandomSpawnOffsetZ = src.RandomSpawnOffsetZ;
        dst.RandomSpawnRotation = src.RandomSpawnRotation;
    }

    public static string GetFileName()
    {
        return global::GameUtils.GetSaveGameDir() + "/RandomSpawnPointData.med";
    }
    public static void WriteVector3(BinaryWriter writer,Vector3 vec)
    {
        writer.Write(vec.x);
        writer.Write(vec.y);
        writer.Write(vec.z);
    }
    public static void ReadVector3(BinaryReader reader, Vector3 vec)
    {
        vec.x = reader.ReadSingle();
        vec.y = reader.ReadSingle();
        vec.z = reader.ReadSingle();
    }
    public static void SetList(List<RandomSpawnPoint> randomSpawnPointList)
    {
        _randomSpawnPointList = randomSpawnPointList;
    }
    public static void LoadData()
    {
        if (!global::Steam.Network.IsClient)
        {
            string fileName = GetFileName();
            if (File.Exists(fileName))
            {

                BinaryReader reader = new BinaryReader(File.OpenRead(fileName));
                ReadBinaryData(reader, ref _randomSpawnPointList);
                reader.Close();
            }
            else
            {
                InitRandomSpawnPointList();
            }
        }
    }
    public static int GetBinaryDataSize()
    {
        return sizeof(int) + _randomSpawnPointList.Count * (sizeof(float) + sizeof(byte));
    }
    public static void SetReadBinaryData(BinaryReader reader)
    {
        ReadBinaryData(reader, ref _randomSpawnPointList);
    }
    public static void ReadBinaryData(BinaryReader reader,ref List<RandomSpawnPoint> pointList)
    {
        pointList = new List<RandomSpawnPoint>();

        int keyCount = reader.ReadInt32();
        for (int i = 0; i < keyCount; i++)
        {
            RandomSpawnPoint point = new RandomSpawnPoint();
            point.Read(reader);
            pointList.Add(point);
        }

    }
    public static void SaveData()
    {
        if (!global::Steam.Network.IsClient)
        {

            string fileName = GetFileName();
            BinaryWriter writer = new BinaryWriter(File.OpenWrite(fileName));
            WriteBinaryData(writer, _randomSpawnPointList);

            writer.Close();
        }
    }
    public static void SetWriteBinaryData(BinaryWriter writer)
    {
        WriteBinaryData(writer, _randomSpawnPointList);
    }
    public static void WriteBinaryData(BinaryWriter writer,List<RandomSpawnPoint> pointList)
    {
        writer.Write(pointList.Count);
        foreach (RandomSpawnPoint point in pointList)
        {
            point.Write(writer);
        }

    }

    private static List<RandomSpawnPoint> _randomSpawnPointList;
    public static void InitRandomSpawnPointList()
    {
        _randomSpawnPointList = new List<RandomSpawnPoint>();
    }
    public static byte GetRotation(RWG2.HubCell.Lot lot,Prefab prefab)
    {
        //
        return (byte)((lot.RoadDirection + prefab.rotationToFaceNorth + prefab.RandomSpawnRotation) % 4);
    }
    public static void AddRandomSpawnPointListHubCell(RWG2.HubCell.Lot lot,RWG2.HubCell hubcell)
    {
        AddRandomSpawnPointList(lot, hubcell.Ruleset.RandamSpawnPrefabFilter);
    }
    public static void AddRandomSpawnPointList(RWG2.HubCell.Lot lot,string filter)
    {
        if (filter != null && lot.PrefabName.Contains(filter))
        {
            Prefab prefab;
            if (lot.Prefab == null)
            {
                prefab = RWG2.RWGCore.Instance.PrefabDecorator.GetPrefab(lot.PrefabName);
            }
            else
            {
                prefab = lot.Prefab;
            }
            byte rotation = (byte)((lot.RoadDirection + prefab.rotationToFaceNorth)% 4);
            float x = lot.Area.center.x;
            float y = lot.Height + prefab.RandomSpawnOffsetY;
            float z = lot.Area.center.y;

            switch (rotation)
            {
                case 0:
                    x += prefab.RandomSpawnOffsetX;
                    z += prefab.RandomSpawnOffsetZ;
                    break;
                case 1:
                    x += prefab.RandomSpawnOffsetZ;
                    z -= prefab.RandomSpawnOffsetX;
                    break;
                case 2:
                    x -= prefab.RandomSpawnOffsetX;
                    z -= prefab.RandomSpawnOffsetZ;
                    break;
                case 3:
                    x -= prefab.RandomSpawnOffsetZ;
                    z += prefab.RandomSpawnOffsetX;
                    break;
            }
            switch (rotation)
            {
                case 1:
                case 3:
                    if(prefab.size.x > prefab.size.z)
                    {
                        float decx = (prefab.size.x - prefab.size.z) / 2;
                        x -= decx;
                    }
                    else
                    {
                        float decz = (prefab.size.z - prefab.size.x) / 2;
                        z -= decz;
                    }

                    break;
            }

            _randomSpawnPointList.Add(new RandomSpawnPoint(x, y, z, GetRotation(lot,prefab)));
        }
    }
    public static float distSqr(Vector2 pointA, Vector2 pointB)
    {
        Vector2 vector = pointA - pointB;
        return vector.x * vector.x + vector.y * vector.y;
    }
    //Prefab.CW
    public static void GetRandomSpawnProperty(Prefab prefab)
    {
        if (prefab.properties.Values.ContainsKey("RandomSpawnPointOffset"))
        {

            string randomSpawnPointOffsetst = prefab.properties.GetStringValue("RandomSpawnPointOffset");
            string[] array = randomSpawnPointOffsetst.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            prefab.RandomSpawnOffsetX = float.Parse(array[0]);
            prefab.RandomSpawnOffsetY = float.Parse(array[1]);
            prefab.RandomSpawnOffsetZ = float.Parse(array[2]);
            prefab.RandomSpawnRotation = byte.Parse(array[3]);
        }
        else
        {
            prefab.RandomSpawnOffsetX = 0.0f;
            prefab.RandomSpawnOffsetY = 0.0f;
            prefab.RandomSpawnOffsetZ = 0.0f;
            prefab.RandomSpawnRotation = 0;

        }

    }
    //Prefab.OW
    public static void GetRandomSpawnDataToProperty(Prefab prefab)
    {
        if (!(prefab.RandomSpawnOffsetX == 0.0f
           && prefab.RandomSpawnOffsetY == 0.0f
           && prefab.RandomSpawnOffsetZ == 0.0f
           && prefab.RandomSpawnRotation == 0)
           )
        {
            prefab.properties.Values["RandomSpawnPointOffset"] = prefab.RandomSpawnOffsetX.ToString() + ","
                + prefab.RandomSpawnOffsetY.ToString() + ","
                + prefab.RandomSpawnOffsetZ.ToString() + ","
                + prefab.RandomSpawnRotation.ToString();
        }

    }
    public static void RulesFromXml(RWG2.Rules.Ruleset ruleset, XmlElement element)
    {
        ruleset.RandamSpawnPrefabFilter = null;
        if (element.HasAttribute("RandamSpawnPrefabFilter"))
        {
            ruleset.RandamSpawnPrefabFilter = element.Attributes["RandamSpawnPrefabFilter"].Value;
        }
    }
    public static SpawnPosition GetRandomSpawnPosition(EntityPlayerLocal entityPlayerLocal)
    {
        if(_randomSpawnPointList.Count > 0)
        {
            RandomSpawnPoint point = _randomSpawnPointList[UnityEngine.Random.Range(0, _randomSpawnPointList.Count)];
            Vector3 rotation = new Vector3(0f, (float)(90 * (point.Rotation & 3)), 0f);
            entityPlayerLocal.SetRotation(rotation);
            SpawnPosition sp = new SpawnPosition(new Vector3(point.CenterX, point.CenterY, point.CenterZ), rotation.y);
            entityPlayerLocal.SetPosition(sp.position);
            return sp;
        }
        else
        {
            return SpawnPosition.Undef;
        }
    }
    public static SpawnPosition GetRandomSpawnPositionNoEntityPlayer( GameManager gameManager)
    {
        if (_randomSpawnPointList.Count > 0)
        {
            RandomSpawnPoint point = _randomSpawnPointList[UnityEngine.Random.Range(0, _randomSpawnPointList.Count)];
            Vector3 rotation = new Vector3(0f, (float)(90 * (point.Rotation & 3)), 0f);
            return new SpawnPosition(new Vector3(point.CenterX, point.CenterY, point.CenterZ), rotation.y);
        }
        else
        {
            return gameManager.GetSpawnPointList().GetRandomSpawnPosition(null, 0, 0);
        }
    }
    public static SpawnPosition GetRandomSpawnPositionNoSet(EntityPlayerLocal entityPlayerLocal, GameManager gameManager)
    {
        if (_randomSpawnPointList.Count > 0)
        {
            RandomSpawnPoint point = _randomSpawnPointList[UnityEngine.Random.Range(0, _randomSpawnPointList.Count)];
            entityPlayerLocal.SetPosition(new Vector3(point.CenterX, point.CenterY, point.CenterZ));
            Vector3 rotation = new Vector3(0f, (float)(90 * (point.Rotation & 3)), 0f);
            entityPlayerLocal.SetRotation(rotation);
            return new SpawnPosition(new Vector3(point.CenterX, point.CenterY, point.CenterZ), rotation.y);
        }
        else
        {
            return gameManager.GetSpawnPointList().GetRandomSpawnPosition(null, 0,0);
        }
    }
    public static void SetAddAttribute(XmlElement element)
    {
        if (element.HasAttribute("NearRandomSpawnPointMax"))
        {
            _nearRandomSpawnPointMaxCount = int.Parse(element.GetAttribute("NearRandomSpawnPointMax"));
        }
        else
        {
            _nearRandomSpawnPointMaxCount = 10;
        }
    }

    private static int _nearRandomSpawnPointMaxCount = 10;
    public static SpawnPosition GetRandomSpawnPositionRange(EntityPlayerLocal entityPlayerLocal, GameManager gameManager)
    {
        return GetRandomSpawnPositionRangeCount(entityPlayerLocal, gameManager, _nearRandomSpawnPointMaxCount);
    }
    public static SpawnPosition GetRandomSpawnPositionRangeCount(EntityPlayerLocal entityPlayerLocal,GameManager gameManager,int rangeCount)
    {

        if (_randomSpawnPointList.Count > 0)
        {
            Vector3 bpPos = entityPlayerLocal.GetDroppedBackpackPosition().ToVector3();
            List<RandomSpawnPointRange> listRange = new List<RandomSpawnPointRange>();

            foreach (RandomSpawnPoint po in _randomSpawnPointList)
            {
                float distX = UnityEngine.Mathf.Abs(po.CenterX - bpPos.x);
                float distY = UnityEngine.Mathf.Abs(po.CenterY - bpPos.y);
                float distZ = UnityEngine.Mathf.Abs(po.CenterZ - bpPos.z);
                float dist = distX * distX + distY * distY + distZ * distZ;
                bool isInsert = false;
                int maxCheck = UnityEngine.Mathf.Min(listRange.Count, rangeCount);
                for(int i = 0;i < maxCheck; i++)
                {
                    if(dist < listRange[i].Dist)
                    {
                        isInsert = true;
                        listRange.Insert(i, new RandomSpawnPointRange(po, dist));
                        break;
                    }
                }
                if (!isInsert && listRange.Count < rangeCount)
                {
                    listRange.Add(new RandomSpawnPointRange(po, dist));
                }
            }


            RandomSpawnPoint point = listRange[UnityEngine.Random.Range(0, listRange.Count)].Point;

            entityPlayerLocal.SetPosition(new Vector3(point.CenterX, point.CenterY, point.CenterZ));
            Vector3 rotation = new Vector3(0f, (float)(90 * (point.Rotation & 3)), 0f);
            entityPlayerLocal.SetRotation(rotation);
            return new SpawnPosition(new Vector3(point.CenterX,point.CenterY,point.CenterZ),0.0f);
        }
        else
        {
            Vector3 vector = entityPlayerLocal.GetDroppedBackpackPosition().ToVector3();
            return gameManager.GetSpawnPointList().GetRandomSpawnPosition(new Vector3?(vector), 500, 2500);
        }
    }
}
